@@ -1,5 +1,5 @@
 /* Spamassassin in local_scan by Marc MERLIN <marc_soft@merlins.org> */
-/* $Id: sa-exim.c,v 1.58 2004/03/10 06:19:20 marcmerlin Exp $ */
+/* $Id: sa-exim.c,v 1.65 2004/08/16 14:37:19 marcmerlin Exp $ */
 /*
 
 The inline comments and minidocs were moved to the distribution tarball
@@ -41,6 +41,10 @@ extern uschar *primary_hostname;
 #define RTLD_NOW 0x002
 #endif    
 
+#ifndef SAFEMESGIDCHARS
+#define SAFEMESGIDCHARS "!#%( )*+,-.0123456789:<=>?@ABCDEFGHIJKLMNOPQRSTUVWXYZ[]^_abcdefghijklmnopqrstuvwxyz{|}~";
+#endif
+
 /* Return the verion of the local_scan ABI, if being compiled as a .so */
 int local_scan_version_major(void)
 {
@@ -176,6 +180,8 @@ static int savemail(int readfd, off_t fdstart, char *dir, char *dirvarname,
     header_line *hl;
     int writefd=0;
     int ret;
+    ssize_t stret;
+    off_t otret;
     char *expand;
     char *fake_env_from;
     int towrite;
@@ -248,9 +254,9 @@ static int savemail(int readfd, off_t fdstart, char *dir, char *dirvarname,
     /* Richard Lithvall made this an option */
     if(SAPrependArchiveWithFrom == 1)
     {
-	fake_env_from=string_sprintf("From %s Thu Jan 1 00:00:01 UTC 1970\n",sender_address);
-	ret=write(writefd, fake_env_from, strlen(fake_env_from));
-	CHECKERR(ret,string_sprintf("'From ' line write in %s", filename),__LINE__);
+	fake_env_from=string_sprintf("From %s Thu Jan  1 00:00:01 1970\n",sender_address);
+	stret=write(writefd, fake_env_from, strlen(fake_env_from));
+	CHECKERR(stret,string_sprintf("'From ' line write in %s", filename),__LINE__);
     }
 
     /* First we need to get the header lines from exim, and then we can read
@@ -264,17 +270,17 @@ static int savemail(int readfd, off_t fdstart, char *dir, char *dirvarname,
 	    hl=hl->next;
 	    continue;
 	}
-	ret=write(writefd,hl->text,strlen(hl->text));
-	CHECKERR(ret,string_sprintf("header line write in %s", filename),__LINE__);
+	stret=write(writefd,hl->text,strlen(hl->text));
+	CHECKERR(stret,string_sprintf("header line write in %s", filename),__LINE__);
 	hl=hl->next;
     }
-    ret=write(writefd,"\n",1);
-    CHECKERR(ret,string_sprintf("header separation write in %s", filename),__LINE__);
+    stret=write(writefd,"\n",1);
+    CHECKERR(stret,string_sprintf("header separation write in %s", filename),__LINE__);
     
     /* Now copy the body to the save file */
     /* we already read from readfd, so we need to reset it */
-    ret=lseek(readfd, fdstart, SEEK_SET);
-    CHECKERR(ret, "lseek reset on spooled message", __LINE__);
+    otret=lseek(readfd, fdstart, SEEK_SET);
+    CHECKERR(otret, "lseek reset on spooled message", __LINE__);
 
     if (SAEximDebug > 8)
     { 
@@ -283,22 +289,23 @@ static int savemail(int readfd, off_t fdstart, char *dir, char *dirvarname,
 
     towrite=SAmaxarchivebody;
     chunk=0;
-    while (towrite>0 && (ret=read(readfd, buffer, MIN(sizeof(buffera), towrite))) > 0)
+    while (towrite>0 && (stret=read(readfd, buffer, MIN(sizeof(buffera), towrite))) > 0)
     {
 	chunk++;
 	if (SAEximDebug > 8)
 	{ 
-	    log_write(0, LOG_MAIN, "SA: Debug9: Processing archive body chunk %d (read %d, and %d can still be written)", chunk, ret, towrite);
+	    log_write(0, LOG_MAIN, "SA: Debug9: Processing archive body chunk %d (read %.0f, and %.0f can still be written)", chunk, (double)stret, (double)towrite);
 	}
-	towrite-=ret;
-	ret=write(writefd, buffer, ret);
-	CHECKERR(ret,string_sprintf("body write in %s", filename),__LINE__);
+	towrite-=stret;
+	stret=write(writefd, buffer, stret);
+	CHECKERR(stret,string_sprintf("body write in %s", filename),__LINE__);
     }
-    CHECKERR(ret, "read body for archival", __LINE__ - 8);
+    CHECKERR(stret, "read body for archival", __LINE__ - 8);
     ret=close(writefd);
     CHECKERR(ret, "Closing spooled message",__LINE__);
     return 0;
 
+    /* catch the global errexit, clean up, and return the error up */
     errexit:
     close(writefd);
     return -1;
@@ -411,8 +418,18 @@ int parsemlheader(char *buffer, FILE *readfh, char *headername, char **header)
 	    {
 		break;
 	    }
-	    /* Slight waste of memory here, oh well... */
-	    *header=string_sprintf("%s\n%s", *header, buffer);
+
+	    /* Guard against humongous header lines */
+	    if (strlen(*header) < 8000)
+	    {
+		/* Slight waste of memory here, oh well... */
+		*header=string_sprintf("%s\n%s", *header, buffer);
+	    }
+	    else
+	    {
+		log_write(0, LOG_MAIN, "SA: Warning: while parsing header %s, ignoring the following trailing line due to header size overflow: %s", headername, buffer);
+
+	    }
 	}
 	if (SAEximDebug > 5)
 	{
@@ -456,7 +473,7 @@ int parsemlheader(char *buffer, FILE *readfh, char *headername, char **header)
 }
 
 
-char *cleanmsgid(char *msgid)
+char *cleanmsgid(char *msgid, char *SAsafemesgidchars)
 {
     char *safemesgid;
     char *ptr;
@@ -468,9 +485,12 @@ char *cleanmsgid(char *msgid)
     /* Clean Message-ID to make sure people can't write on our FS */
     while (*ptr)
     {
-	if (*ptr == '/')
+	/* This might be more aggressive than you want, but since you
+	 * potentially have shell programs dealing with the resulting filenames
+	 * let's make it a bit safer */
+	if (strchr(SAsafemesgidchars, *ptr) == NULL)
 	{
-	    *ptr='-';
+	    *ptr='_';
 	}
 	ptr++;
     }
@@ -490,6 +510,7 @@ int local_scan(volatile int fd, uschar **return_text)
 {
 #warning you should not worry about the "might be clobbered by longjmp", see source
     int ret;
+    ssize_t stret;
     int pid;
     int writefd[2];
     int readfd[2];
@@ -507,7 +528,7 @@ int local_scan(volatile int fd, uschar **return_text)
     int isspam=0;
     int gotsa=0;
     int chunk;
-    int towrite;
+    off_t towrite;
     char *mailinfo;
     float spamvalue=0.0;
     char *spamstatus=NULL;
@@ -526,7 +547,10 @@ int local_scan(volatile int fd, uschar **return_text)
 
     /* Options we read from /etc/exim4/sa-exim.conf */
     static char *SAspamcpath=SPAMC_LOCATION;
+    static char *SAsafemesgidchars=SAFEMESGIDCHARS
     static char *SAspamcSockPath=NULL;
+    static char *SAspamcPort="783";
+    static char *SAspamcHost="127.0.0.1";
     static char *SAEximRunCond="0";
     static char *SAEximRejCond="1";
     static int SAmaxbody=250*1024;
@@ -578,8 +602,12 @@ int local_scan(volatile int fd, uschar **return_text)
 
 
     /* Any error can write the faulty message to mesgfn, so we need to
-       give it a value right now. We'll replace it with the real Message-ID later */
-    safemesgid=cleanmsgid(message_id);
+       give it a value right now. We'll set the real value later */
+    /* message_id here comes from Exim, it's an internal disk Mesg-Id format
+       which doesn't correlate to the actual message's Mesg-Id. We shouldn't
+       need to clean it, and besides, SAsafemesgidchars hasn't been read from
+       the config file yet, but eh, safety is always a good thing, right? */
+    safemesgid=cleanmsgid(message_id, SAsafemesgidchars);
     mesgfn=string_sprintf("%d_%s", time(NULL), safemesgid);
 
     /* We won't scan local messages. I think exim bypasses local_scan for a
@@ -591,6 +619,14 @@ int local_scan(volatile int fd, uschar **return_text)
     } 
     */
 
+    /* If you discard a mail with exim ACLs, we get 0 recipients, so let's just
+     * accept the mail, which is bogus either way since it'll get dropped
+     * (thanks to John Horne for reporting this corner case) */
+    if (recipients_count == 0) 
+    {
+	return LOCAL_SCAN_ACCEPT;
+    }
+
     /* 
      * We keep track of whether we've alrady read the config file, but since
      * exim spawns itself, it will get read by exim children even though you
@@ -666,7 +702,10 @@ int local_scan(volatile int fd, uschar **return_text)
 
 	    M_CHECKFORVAR(SAEximDebug, "%d");
 	    M_CHECKFORSTR(SAspamcpath);
+	    M_CHECKFORSTR(SAsafemesgidchars);
 	    M_CHECKFORSTR(SAspamcSockPath);
+	    M_CHECKFORSTR(SAspamcPort);
+	    M_CHECKFORSTR(SAspamcHost);
 	    M_CHECKFORSTR(SAEximRunCond);
 	    M_CHECKFORSTR(SAEximRejCond);
 	    M_CHECKFORVAR(SAmaxbody, "%d");
@@ -755,7 +794,7 @@ int local_scan(volatile int fd, uschar **return_text)
 
     if (SAEximDebug > 3)
     {
-	log_write(0, LOG_MAIN, "SA: Debug4: Message body is about %d bytes and the initial offset is %d", fdsize-18, fdstart);
+	log_write(0, LOG_MAIN, "SA: Debug4: Message body is about %.0f bytes and the initial offset is %.0f", (double)(fdsize-18), (double)fdstart);
     }
 
     if (fdsize > SAmaxbody)
@@ -775,7 +814,7 @@ int local_scan(volatile int fd, uschar **return_text)
 
 	    if (expand[0] == 0 || (expand[0] == '0' && expand[1] == 0))
 	    {
-		log_write(0, LOG_MAIN, "SA: Action: check skipped due to message size (%d bytes) and SATruncBodyCond expanded to false (Message-Id: %s). %s", fdsize-18, safemesgid, mailinfo);
+		log_write(0, LOG_MAIN, "SA: Action: check skipped due to message size (%.0f bytes) and SATruncBodyCond expanded to false (Message-Id: %s). %s", (double)(fdsize-18), safemesgid, mailinfo);
 		header_add(' ', "X-SA-Exim-Scanned: No (on %s); Message bigger than SAmaxbody (%d)\n", primary_hostname, SAmaxbody);
 		return LOCAL_SCAN_ACCEPT;
 	    }
@@ -783,12 +822,12 @@ int local_scan(volatile int fd, uschar **return_text)
 
 	if (SAEximDebug > 1)
 	{
-	    log_write(0, LOG_MAIN, "SA: Debug2: Message body is about %d bytes and SATruncBodyCond expanded to true, will feed a truncated body to SA", fdsize-18);
+	    log_write(0, LOG_MAIN, "SA: Debug2: Message body is about %.0f bytes and SATruncBodyCond expanded to true, will feed a truncated body to SA", (double)(fdsize-18));
 	}
 
 	/* Let's feed exactly spamc will accept */
 	scansize=SAmaxbody;
-	header_add(' ', "X-SA-Exim-Scan-Truncated: Fed %d bytes of the body to SA instead of %d\n", scansize, fdsize);
+	header_add(' ', "X-SA-Exim-Scan-Truncated: Fed %.0f bytes of the body to SA instead of %.0f\n", (double)scansize, (double)fdsize);
     }
     else
     {
@@ -856,7 +895,8 @@ int local_scan(volatile int fd, uschar **return_text)
          * I could implement the spamc protocol and talk to spamd directly
          * instead of forking spamc, but considering the overhead spent
          * in spamd, forking off spamc seemed acceptable rather than
-         * re-implementing and tracking the spamc/spamd protocol
+         * re-implementing and tracking the spamc/spamd protocol or linking
+	 * with a possibly changing library
 	 */
 	/* Ok, we cheat, spamc cares about how big the whole message is and
          * we only know about the body size, so I'll  give an extra 16K
@@ -868,9 +908,10 @@ int local_scan(volatile int fd, uschar **return_text)
 	}
 	else
 	{
-	    ret=execl(SAspamcpath, "spamc", "-s", string_sprintf("%d", SAmaxbody+16384), NULL);
+	    ret=execl(SAspamcpath, "spamc", "-s", string_sprintf("%d", SAmaxbody+16384), "-d", SAspamcHost, "-p", SAspamcPort, NULL);
 	    CHECKERR(ret,string_sprintf("exec %s", SAspamcpath),__LINE__);
 	}
+	
     }
 
     if (SAEximDebug > 8)
@@ -902,13 +943,13 @@ int local_scan(volatile int fd, uschar **return_text)
 	    continue;
 	}
 
-	ret=write(writefd[1],hl->text,strlen(hl->text));
-	CHECKERR(ret,"header line write",__LINE__);
+	stret=write(writefd[1],hl->text,strlen(hl->text));
+	CHECKERR(stret,"header line write",__LINE__);
 
 	hl=hl->next;
     }
-    ret=write(writefd[1],"\n",1);
-    CHECKERR(ret,"header separation write",__LINE__);
+    stret=write(writefd[1],"\n",1);
+    CHECKERR(stret,"header separation write",__LINE__);
 
     if (SAEximDebug > 6)
     {
@@ -917,22 +958,22 @@ int local_scan(volatile int fd, uschar **return_text)
 
     towrite=scansize;
     chunk=0;
-    while (towrite>0 && (ret=read(fd, buffer, MIN(sizeof(buffera), towrite))) > 0)
+    while (towrite>0 && (stret=read(fd, buffer, MIN(sizeof(buffera), towrite))) > 0)
     {
 	chunk++;
 	if (SAEximDebug > 8)
 	{ 
-	    log_write(0, LOG_MAIN, "SA: Debug9: spamc body going to write chunk %d (read %d, %d left to write)", chunk, ret, towrite);
+	    log_write(0, LOG_MAIN, "SA: Debug9: spamc body going to write chunk %d (read %.0f, %.0f left to write)", chunk, (double)stret, (double)towrite);
 	}
-	towrite-=ret;
-	ret=write(writefd[1], buffer, ret);
-	CHECKERR(ret,"body write in",__LINE__);
+	towrite-=stret;
+	stret=write(writefd[1], buffer, stret);
+	CHECKERR(stret,"body write in",__LINE__);
 	if (SAEximDebug > 8)
 	{ 
-	    log_write(0, LOG_MAIN, "SA: Debug9: Spamc body wrote chunk %d (wrote %d, %d left to write)", chunk, ret, towrite);
+	    log_write(0, LOG_MAIN, "SA: Debug9: Spamc body wrote chunk %d (wrote %.0f, %.0f left to write)", chunk, (double)stret, (double)towrite);
 	}
     }
-    CHECKERR(ret, "read body", __LINE__ - 14);
+    CHECKERR(stret, "read body", __LINE__ - 14);
     close(writefd[1]);
 
     if (SAEximDebug > 5)
@@ -1038,6 +1079,12 @@ restart:
 	    }
 
 	    start=strstr(xspamstatus, "hits=");
+	    /* Support SA 3.0 format */
+	    if (start == NULL)
+	    {
+	    	start=strstr(xspamstatus, "score=");
+	    }
+
 	    end=strstr(xspamstatus, " tests=");
 	    if (end == NULL)
 	    {
@@ -1095,8 +1142,8 @@ restart:
 	    char *end;
 	    char *mesgid=NULL;
 
-	    start=index(buffer, '<');
-	    end=index(buffer, '>');
+	    start=strchr(buffer, '<');
+	    end=strchr(buffer, '>');
 
 	    if (start == NULL || end == NULL)
 	    {
@@ -1111,7 +1158,7 @@ restart:
                 /* We replace the exim Message-ID with the one read from
                 the message * as we use this to detect dupes when we
                 send 45x and get the same * message multiple times */
-		safemesgid=cleanmsgid(mesgid);
+		safemesgid=cleanmsgid(mesgid, SAsafemesgidchars);
 		mesgfn=string_sprintf("%d_%s", time(NULL), safemesgid);
 
 		if (SAEximDebug > 5)
@@ -1147,8 +1194,8 @@ restart:
 		log_write(0, LOG_MAIN, "SA: Debug9: Read body from SA; line %d (read %d)", line, strlen(buffer));
 	    }
 
-	    ret=write(fd, buffer, strlen(buffer));
-	    CHECKERR(ret,string_sprintf("SA body write to msg"),__LINE__);
+	    stret=write(fd, buffer, strlen(buffer));
+	    CHECKERR(stret,string_sprintf("SA body write to msg"),__LINE__);
 	    if (SAEximDebug > 8)
 	    {
 		log_write(0, LOG_MAIN, "SA: Debug9: Wrote to msg; line %d (wrote %d)", line, ret);
